-- 
-- Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
--
-- 
-- Filename: clock_generator.vhd
--
-- Function: generate various clocks for the DIY calculator CPU
-- 
--
-- Changelog
--
--  1.0  27.Jul.2006   JH   new
--  1.1   1.Sep.2006   JH   accurate frequency in favour of symm. duty cycle
--  1.2  18.Sep.2006   JH   can write CPU clk divider
--  1.3  25.Sep.2006   JH   add SPI controller
--        4.Oct.2006   JH   fine tune divider algorithm
--
--


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity clock_generator is
  port (
    rst      : in std_logic;
    clk_in   : in std_logic;   -- input clock (25MHz)
    sel      : in std_logic;
    nWR      : in std_logic;
    ddrive   : in std_logic;
    d_in     : in std_logic_vector (7 downto 0);
    d_out    : out std_logic_vector (7 downto 0);
    cpu_clk  : out std_logic;  -- CPU clock (3Hz .. 12.5MHz)
    uart_clk : out std_logic;  -- UART clock (115.2kHz:38400b/sec) 
    lcd_clk  : out std_logic;  -- LCD controller clock (1MHz)
    key_clk  : out std_logic;  -- keypad controller clock (10kHz)
    spi_clk  : out std_logic   -- SPI controller clock (6.25MHz)
  );
end clock_generator;

architecture divider of clock_generator is
  constant SYS_FREQ  : integer :=  25000000;  -- 25MHz
  constant UART_FREQ : integer :=    115200;  -- 3*38.4kHz
  constant LCD_FREQ  : integer :=   1000000;  -- 1MHz
  constant KEY_FREQ  : integer :=     10000;  -- 10kHz
  constant SPI_FREQ  : integer :=   6250000;  -- 6.25MHz
  signal uart_cnt : integer;
  signal lcd_cnt  : integer;
  signal key_cnt  : integer;
  signal spi_cnt  : integer;
  signal cpu_cnt, cpu_end : std_logic_vector (23 downto 0);
  signal cpu_div : std_logic_vector (7 downto 0);
  signal wr_cpu_div : std_logic;
begin
  -- write the clock divider
  wr_cpu_div <= not nWR and sel;
  cpu_div_latch : process (rst, wr_cpu_div)
  begin
    if rst = '1' then
      cpu_div <= x"01";
    elsif falling_edge (wr_cpu_div) then
      if d_in = x"00" then
        cpu_div <= x"01";
      else
        cpu_div <= d_in;
      end if;
    end if;
  end process;

  -- read the clock divider
  d_out(0) <= sel and not ddrive and cpu_div(0);
  d_out(1) <= sel and not ddrive and cpu_div(1);
  d_out(2) <= sel and not ddrive and cpu_div(2);
  d_out(3) <= sel and not ddrive and cpu_div(3);
  d_out(4) <= sel and not ddrive and cpu_div(4);
  d_out(5) <= sel and not ddrive and cpu_div(5);
  d_out(6) <= sel and not ddrive and cpu_div(6);
  d_out(7) <= sel and not ddrive and cpu_div(7);

  -- the counter
    cpu_divider : process (rst, clk_in)
    variable segment : std_logic_vector (15 downto 0);
  begin
    if rst = '1' then
      cpu_cnt <= x"000000";
      cpu_clk <= '0';
    elsif rising_edge (clk_in) then
      if unsigned(cpu_cnt) < unsigned('0' & cpu_end(23 downto 1)) then
        cpu_clk <= '0';
      else
        cpu_clk <= '1';
      end if;
      if unsigned(cpu_cnt) >= unsigned(cpu_end) then
        cpu_clk <= '0';
        cpu_cnt <= x"000001";
      else
        cpu_cnt <= cpu_cnt + 1;
      end if;
    end if;
    case cpu_div (7 downto 4) is
      when x"0" => -- 0000 0000 0000 0000 000x xxx0
        segment := x"0000";
        cpu_end <= "000" & segment & cpu_div(3 downto 0) & '0';

      when x"1" => -- 0000 0000 0000 0000 001x xxx0
        segment := x"0001";
        cpu_end <= "000" & segment & cpu_div(3 downto 0) & '0';

      when x"2" => -- 0000 0000 0000 0000 010x xxx0
        segment := x"0002";
        cpu_end <= "000" & segment & cpu_div(3 downto 0) & '0';

      when x"3" => -- 0000 0000 0000 0000 100x xxx0
        segment := x"0004";
        cpu_end <= "000" & segment & cpu_div(3 downto 0) & '0';

      when x"4" => -- 0000 0000 0000 0010 000x xxx0
        segment := x"0008";
        cpu_end <= "00" & segment & '0' & cpu_div(3 downto 0) & '0';

      when x"5" => -- 0000 0000 0000 0100 00xx xx00
        segment := x"0010";
        cpu_end <= "00" & segment & cpu_div(3 downto 0) & "00";

      when x"6" => -- 0000 0000 0001 0000 00xx xx00
        segment := x"0020";
        cpu_end <= '0' & segment & '0' & cpu_div(3 downto 0) & "00";

      when x"7" => -- 0000 0000 0010 0000 00xx xx00
        segment := x"0040";
        cpu_end <= '0' & segment & '0' & cpu_div(3 downto 0) & "00";

      when x"8" => -- 0000 0000 0100 0000 0xxx x000
        segment := x"0080";
        cpu_end <= '0' & segment & cpu_div(3 downto 0) & "000";

      when x"9" => -- 0000 0001 0000 0000 0xxx x000
        segment := x"0100";
        cpu_end <= segment & '0' & cpu_div(3 downto 0) & "000";

      when x"A" => -- 0000 0010 0000 0000 0xxx x000
        segment := x"0200";
        cpu_end <= segment & '0' & cpu_div(3 downto 0) & "000";

      when x"B" => -- 0000 0100 0000 0000 xxxx 0000
        segment := x"0400";
        cpu_end <= segment & cpu_div(3 downto 0) & "0000";

      when x"C" => -- 0001 0000 0000 0000 xxxx 0000
        segment := x"0800";
        cpu_end <= segment(14 downto 0) & '0' & cpu_div(3 downto 0) & "0000";

      when x"D" => -- 0010 0000 0000 0000 xxxx 0000
        segment := x"1000";
        cpu_end <= segment(14 downto 0) & '0' & cpu_div(3 downto 0) & "0000";

      when x"E" => -- 0100 0000 0000 000x xxx0 0000
        segment := x"2000";
        cpu_end <= segment(14 downto 0) & cpu_div(3 downto 0) & "00000";

      when x"F" => -- 1000 0000 0000 000x xxx0 0000
        segment := x"4000";
        cpu_end <= segment(14 downto 0) & cpu_div(3 downto 0) & "00000";
      when others => null;
    end case;
  end process;

  uart_divider : process (rst, clk_in)
    begin
    if rst = '1' then
      uart_cnt <= 0;
      uart_clk <= '0';
    elsif rising_edge (clk_in) then
      uart_cnt <= uart_cnt + 1;
      if uart_cnt >= ((SYS_FREQ/2) / UART_FREQ) then
        uart_clk <= '0';
      else
        uart_clk <= '1';
      end if;
      if uart_cnt >= ((SYS_FREQ / UART_FREQ) - 1) then
        uart_cnt <= 0;
      end if;
    end if;
  end process;

  lcd_divider : process (rst, clk_in)
    begin
    if rst = '1' then
      lcd_cnt <= 0;
      lcd_clk <= '0';
    elsif rising_edge (clk_in) then
      if lcd_cnt >= ((SYS_FREQ/2) / LCD_FREQ) then
        lcd_clk <= '0';
      else
        lcd_clk <= '1';
      end if;
      if lcd_cnt >= ((SYS_FREQ / LCD_FREQ) - 1) then
        lcd_cnt <= 0;
      else
        lcd_cnt <= lcd_cnt + 1;
      end if;
    end if;
  end process;

  key_divider : process (rst, clk_in)
    begin
    if rst = '1' then
      key_cnt <= 0;
      key_clk <= '0';
    elsif rising_edge (clk_in) then
      if key_cnt >= ((SYS_FREQ/2) / KEY_FREQ) then
        key_clk <= '0';
      else
        key_clk <= '1';
      end if;
      if key_cnt >= ((SYS_FREQ / KEY_FREQ) - 1) then
        key_cnt <= 0;
      else
        key_cnt <= key_cnt + 1;
      end if;
    end if;
  end process;

  spi_divider : process (rst, clk_in)
    begin
    if rst = '1' then
      spi_cnt <= 0;
      spi_clk <= '0';
    elsif rising_edge (clk_in) then
      if spi_cnt >= ((SYS_FREQ/2) / SPI_FREQ) then
        spi_clk <= '0';
      else
        spi_clk <= '1';
      end if;
      if spi_cnt >= ((SYS_FREQ / SPI_FREQ) - 1) then
        spi_cnt <= 0;
      else
        spi_cnt <= spi_cnt + 1;
      end if;
    end if;
  end process;
end divider;
